# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("any",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._connect.numpy import UNSUPPORTED
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def any(
    array,
    axis=None,
    *,
    keepdims=False,
    mask_identity=False,
    highlevel=True,
    behavior=None,
):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns True in each group of elements from `array` (many types supported,
    including all Awkward Arrays and Records) if any values are True; False
    otherwise. Thus, it represents reduction over the "logical or" operation,
    whose identity is False (i.e. asking if there are any True values in an
    empty list results in False). This operation is the same as NumPy's
    [any](https://docs.scipy.org/doc/numpy/reference/generated/numpy.any.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, keepdims, mask_identity, highlevel, behavior)


def _impl(array, axis, keepdims, mask_identity, highlevel, behavior):
    axis = regularize_axis(axis)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    behavior = behavior_of(array, behavior=behavior)
    reducer = ak._reducers.Any()

    out = ak._do.reduce(
        layout,
        reducer,
        axis=axis,
        mask=mask_identity,
        keepdims=keepdims,
        behavior=behavior,
    )
    if isinstance(out, (ak.contents.Content, ak.record.Record)):
        return wrap_layout(out, behavior, highlevel)
    else:
        return out


@ak._connect.numpy.implements("any")
def _nep_18_impl(a, axis=None, out=UNSUPPORTED, keepdims=False, *, where=UNSUPPORTED):
    return any(a, axis=axis, keepdims=keepdims)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("ptp",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._connect.numpy import UNSUPPORTED
from awkward._dispatch import high_level_function
from awkward._layout import maybe_posaxis
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import regularize_axis

np = NumpyMetadata.instance()


@high_level_function
def ptp(array, axis=None, *, keepdims=False, mask_identity=True):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        axis (None or int): If None, combine all values from the array into
            a single scalar result; if an int, group by that axis: `0` is the
            outermost, `1` is the first level of nested lists, etc., and
            negative `axis` counts from the innermost: `-1` is the innermost,
            `-2` is the next level up, etc.
        keepdims (bool): If False, this reducer decreases the number of
            dimensions by 1; if True, the reduced values are wrapped in a new
            length-1 dimension so that the result of this operation may be
            broadcasted with the original array.
        mask_identity (bool): If True, reducing over empty lists results in
            None (an option type); otherwise, reducing over empty lists
            results in the operation's identity of 0.

    Returns the range of values in each group of elements from `array` (many
    types supported, including all Awkward Arrays and Records). The range of
    an empty list is None, unless `mask_identity=False`, in which case it is 0.
    This operation is the same as NumPy's
    [ptp](https://docs.scipy.org/doc/numpy/reference/generated/numpy.ptp.html)
    if all lists at a given dimension have the same length and no None values,
    but it generalizes to cases where they do not.

    For example, with

        >>> array = ak.Array([[0, 1, 2, 3],
        ...                   [          ],
        ...                   [4, 5      ]])

    The range of the innermost lists is

        >>> ak.ptp(array, axis=-1)
        <Array [3, None, 1] type='3 * ?int64'>

    because there are three lists, the first has a range of `3`, the second is
    `None` because the list is empty, and the third has a range of `1`. Similarly,

        >>> ak.ptp(array, axis=-1, mask_identity=False)
        <Array [3, 0, 1] type='3 * float64'>

    The second value is `0` because the list is empty.

    See #ak.sum for a more complete description of nested list and missing
    value (None) handling in reducers.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, axis, keepdims, mask_identity)


def _impl(array, axis, keepdims, mask_identity):
    axis = regularize_axis(axis)
    behavior = behavior_of(array)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)

    with np.errstate(invalid="ignore", divide="ignore"):
        maxi = ak.operations.ak_max._impl(
            layout,
            axis,
            True,
            None,
            mask_identity,
            highlevel=True,
            behavior=behavior,
        )
        mini = ak.operations.ak_min._impl(
            layout,
            axis,
            True,
            None,
            True,
            highlevel=True,
            behavior=behavior,
        )
        out = maxi - mini

        # Check that removed code was not needed!
        assert maxi is not None and mini is not None

        if not mask_identity:
            out = ak.highlevel.Array(ak.operations.fill_none(out, 0, axis=-1))

        if axis is None:
            if not keepdims:
                out = out[(0,) * out.ndim]
        else:
            if not keepdims:
                posaxis = maybe_posaxis(out.layout, axis, 1)
                out = out[(slice(None, None),) * posaxis + (0,)]

        return out


@ak._connect.numpy.implements("ptp")
def _nep_18_impl(a, axis=None, out=UNSUPPORTED, keepdims=False):
    return ptp(a, axis=axis, keepdims=keepdims)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("with_field",)
import copy

import awkward as ak
from awkward._backends.dispatch import backend_of
from awkward._backends.numpy import NumpyBackend
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata
from awkward._regularize import is_non_string_like_sequence

np = NumpyMetadata.instance()


cpu = NumpyBackend.instance()


@high_level_function
def with_field(array, what, where=None, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        what: Array-like data (anything #ak.to_layout recognizes) to add as a new field.
        where (None or str or non-empy sequence of str): If None, the new field
            has no name (can be accessed as an integer slot number in a
            string); If str, the name of the new field. If a sequence, it is
            interpreted as a path where to add the field in a nested record.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Returns an #ak.Array or #ak.Record (or low-level equivalent, if
    `highlevel=False`) with a new field attached. This function does not
    change the array in-place.

    See #ak.Array.__setitem__ and #ak.Record.__setitem__ for a variant that
    changes the high-level object in-place. (These methods internally use
    #ak.with_field, so performance is not a factor in choosing one over the
    other.)
    """
    # Dispatch
    yield array, what

    # Implementation
    return _impl(array, what, where, highlevel, behavior)


def _impl(base, what, where, highlevel, behavior):
    if not (
        where is None
        or isinstance(where, str)
        or (
            is_non_string_like_sequence(where)
            and all(isinstance(x, str) for x in where)
        )
    ):
        raise TypeError(
            "New fields may only be assigned by field name(s) "
            "or as a new integer slot by passing None for 'where'"
        )

    if is_non_string_like_sequence(where) and len(where) > 1:
        return _impl(
            base,
            _impl(
                base[where[0]],
                what,
                where[1:],
                highlevel,
                behavior,
            ),
            where[0],
            highlevel,
            behavior,
        )
    else:
        # If we have an iterable here, pull out the only ti
        if is_non_string_like_sequence(where):
            where = where[0]

        behavior = behavior_of(base, what, behavior=behavior)
        backend = backend_of(base, what, default=cpu)

        base = ak.operations.to_layout(
            base, allow_record=True, allow_other=False
        ).to_backend(backend)
        what = ak.operations.to_layout(what, allow_record=True, allow_other=True)
        if isinstance(what, (ak.contents.Content, ak.record.Record)):
            what = what.to_backend(backend)

        keys = copy.copy(base.fields)
        if where in base.fields:
            keys.remove(where)

        def purelist_is_record(layout):
            result = False

            def action_is_record(input, **kwargs):
                nonlocal result

                if input.is_record:
                    result = True
                    return input
                elif input.is_union:
                    result = all(purelist_is_record(x) for x in input.contents)
                    return input
                else:
                    return None

            ak._do.recursively_apply(layout, action_is_record, return_array=False)
            return result

        if not purelist_is_record(base):
            raise ValueError("no tuples or records in array; cannot add a new field")

        def action(inputs, **kwargs):
            base, what = inputs
            backend = base.backend

            if isinstance(base, ak.contents.RecordArray):
                if what is None:
                    what = ak.contents.IndexedOptionArray(
                        ak.index.Index64(
                            backend.index_nplike.full(base.length, -1, dtype=np.int64),
                            nplike=backend.index_nplike,
                        ),
                        ak.contents.EmptyArray(),
                    )
                elif not isinstance(what, ak.contents.Content):
                    what = ak.contents.NumpyArray(
                        backend.nplike.repeat(
                            backend.nplike.asarray(what),
                            backend.nplike.shape_item_as_index(base.length),
                        )
                    )
                if base.is_tuple:
                    # Preserve tuple-ness
                    if where is None:
                        fields = None
                    # Otherwise the tuple becomes a record
                    else:
                        fields = [*keys, where]
                # Records with `where=None` will create a tuple-like key
                elif where is None:
                    fields = [*keys, str(len(keys))]
                else:
                    fields = [*keys, where]
                out = ak.contents.RecordArray(
                    [base[k] for k in keys] + [what],
                    fields,
                    parameters=base.parameters,
                )
                return (out,)
            else:
                return None

        out = ak._broadcasting.broadcast_and_apply(
            [base, what],
            action,
            behavior,
            right_broadcast=False,
        )

        assert isinstance(out, tuple) and len(out) == 1

        return wrap_layout(out[0], behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    array = ak.Array(
        ak.contents.ListOffsetArray(
            ak.index.Index64(
                np.array(
                    [0, 2, 4, 4],
                    dtype=np.int_,
                )
            ),
            ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(4 * 2)), size=2),
        )
    )

    assert ak.sum(array, axis=1).to_list() == [[2, 4], [10, 12], [0, 0]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_values_astype_datetime():
    array1 = ak.operations.values_astype(
        ak.highlevel.Array([1567416600000]), "datetime64[ms]"
    )
    assert str(array1.type) == "1 * datetime64[ms]"
    assert array1 == [np.datetime64("2019-09-02T09:30:00")]

    array2 = ak.operations.values_astype(
        ak.highlevel.Array([1567416600000]), np.dtype("M8[ms]")
    )
    assert str(array2.type) == "1 * datetime64[ms]"
    assert array2 == [np.datetime64("2019-09-02T09:30:00")]

    array3 = ak.operations.values_astype(
        ak.highlevel.Array([1567416600000000, None]), "datetime64[us]"
    )
    assert to_list(array3) == [np.datetime64("2019-09-02T09:30:00"), None]


def test_modulo_units():
    array1 = ak.operations.values_astype(
        ak.highlevel.Array([1]), np.dtype("datetime64[100as/1]")
    )
    assert array1 == [np.datetime64("1970-01-01T00:00:00.000000000000000100", "100as")]

    array2 = ak.operations.values_astype(
        ak.highlevel.Array([1]), np.dtype("datetime64[10s/2]")
    )
    assert array2 == [np.datetime64("1970-01-01T00:00:05.000", "5000ms")]


def test_float_values_astype_datetime():
    array = ak.highlevel.Array([1.9999, 1567416600000, 0, None, 11, 0.555])
    assert str(array.type) == "6 * ?float64"

    dt_array = ak.operations.values_astype(array, "datetime64[ms]")
    assert str(dt_array.type) == "6 * ?datetime64[ms]"
    assert dt_array.to_list() == [
        np.datetime64("1970-01-01T00:00:00.001"),
        np.datetime64("2019-09-02T09:30:00.000"),
        np.datetime64("1970-01-01T00:00:00.000"),
        None,
        np.datetime64("1970-01-01T00:00:00.011"),
        np.datetime64("1970-01-01T00:00:00.000"),
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("nan_to_none",)
import awkward as ak
from awkward._behavior import behavior_of
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def nan_to_none(array, *, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts NaN ("not a number") into None, i.e. missing values with option-type.

    See also #ak.nan_to_num to convert NaN or infinity to specified values.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, highlevel, behavior)


def _impl(array, highlevel, behavior):
    def action(layout, continuation, **kwargs):
        if isinstance(layout, ak.contents.NumpyArray) and issubclass(
            layout.dtype.type, np.floating
        ):
            mask = layout.backend.nplike.isnan(layout.data)
            return ak.contents.ByteMaskedArray(
                ak.index.Index8(mask, nplike=layout.backend.index_nplike),
                layout,
                valid_when=False,
            )

        elif (layout.is_option or layout.is_indexed) and (
            isinstance(layout.content, ak.contents.NumpyArray)
            and issubclass(layout.content.dtype.type, np.floating)
        ):
            return continuation()

        else:
            return None

    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    behavior = behavior_of(array, behavior=behavior)
    out = ak._do.recursively_apply(layout, action, behavior)
    return wrap_layout(out, behavior, highlevel)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
from __future__ import annotations

from awkward._errors import deprecate
from awkward._typing import Self, final
from awkward._util import UNSET
from awkward.types.type import Type


@final
class UnknownType(Type):
    def copy(self, *, parameters=UNSET, typestr=UNSET) -> Self:
        if not (parameters is UNSET or parameters is None or len(parameters) == 0):
            raise TypeError(f"{type(self).__name__} cannot contain parameters")
        return UnknownType(
            typestr=self._typestr if typestr is UNSET else typestr,
        )

    def __init__(self, *, parameters=None, typestr=None):
        if not (parameters is None or len(parameters) == 0):
            raise TypeError(f"{type(self).__name__} cannot contain parameters")
        if typestr is not None and not isinstance(typestr, str):
            raise TypeError(
                "{} 'typestr' must be of type string or None, not {}".format(
                    type(self).__name__, repr(typestr)
                )
            )
        self._parameters = None
        self._typestr = typestr

    def _str(self, indent, compact, behavior):
        if self._typestr is not None:
            deprecate("typestr argument is deprecated", "2.4.0")

        typestr = self._typestr
        if typestr is not None:
            out = [typestr]

        else:
            params = self._str_parameters()
            if params is None:
                out = ["unknown"]
            else:
                out = ["unknown[", params, "]"]

        return [self._str_categorical_begin(), *out, self._str_categorical_end()]

    def __repr__(self):
        args = self._repr_args()
        return "{}({})".format(type(self).__name__, ", ".join(args))

    def _is_equal_to(self, other, all_parameters: bool):
        return isinstance(other, type(self))



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


def test_one_level():
    base = ak.operations.zip(
        {"a": ak.operations.zip({"x": [1, 2, 3]}), "b": [1, 2, 3]},
        depth_limit=1,
    )
    what = ak.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(ak.operations.with_field(base, what, where=["a", "y"])) == [
        {"b": 1, "a": {"x": 1, "y": 1.1}},
        {"b": 2, "a": {"x": 2, "y": 2.2}},
        {"b": 3, "a": {"x": 3, "y": 3.3}},
    ]
    assert ak.operations.fields(base) == ["a", "b"]

    base["a", "y"] = what
    assert to_list(base) == [
        {"b": 1, "a": {"x": 1, "y": 1.1}},
        {"b": 2, "a": {"x": 2, "y": 2.2}},
        {"b": 3, "a": {"x": 3, "y": 3.3}},
    ]


def test_two_level():
    base = ak.operations.zip(
        {
            "A": ak.operations.zip(
                {"a": ak.operations.zip({"x": [1, 2, 3]}), "b": [1, 2, 3]}
            ),
            "B": [1, 2, 3],
        },
        depth_limit=1,
    )
    what = ak.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(ak.operations.with_field(base, what, where=["A", "a", "y"])) == [
        {"B": 1, "A": {"b": 1, "a": {"x": 1, "y": 1.1}}},
        {"B": 2, "A": {"b": 2, "a": {"x": 2, "y": 2.2}}},
        {"B": 3, "A": {"b": 3, "a": {"x": 3, "y": 3.3}}},
    ]
    assert ak.operations.fields(base) == ["A", "B"]

    base["A", "a", "y"] = what
    assert to_list(base) == [
        {"B": 1, "A": {"b": 1, "a": {"x": 1, "y": 1.1}}},
        {"B": 2, "A": {"b": 2, "a": {"x": 2, "y": 2.2}}},
        {"B": 3, "A": {"b": 3, "a": {"x": 3, "y": 3.3}}},
    ]


def test_replace_the_only_field():
    base = ak.operations.zip({"a": ak.operations.zip({"x": [1, 2, 3]})}, depth_limit=1)
    what = ak.Array([1.1, 2.2, 3.3], check_valid=True)
    assert to_list(ak.operations.with_field(base, what, where=["a", "y"])) == [
        {"a": {"x": 1, "y": 1.1}},
        {"a": {"x": 2, "y": 2.2}},
        {"a": {"x": 3, "y": 3.3}},
    ]
    assert ak.operations.fields(base) == ["a"]

    base["a", "y"] = what
    assert to_list(base) == [
        {"a": {"x": 1, "y": 1.1}},
        {"a": {"x": 2, "y": 2.2}},
        {"a": {"x": 3, "y": 3.3}},
    ]


def test_check_no_field():
    base = ak.operations.zip({"a": ak.operations.zip({"x": [1, 2, 3]})}, depth_limit=1)
    what = ak.Array([1.1, 2.2, 3.3], check_valid=True)

    assert to_list(ak.operations.with_field(base, what)) == [
        {"a": {"x": 1}, "1": 1.1},
        {"a": {"x": 2}, "1": 2.2},
        {"a": {"x": 3}, "1": 3.3},
    ]

    with pytest.raises(ValueError):
        ak.operations.with_field(what, what)

    content1 = ak.contents.NumpyArray(np.array([1, 2, 3]))
    recordarray = ak.contents.RecordArray([content1], None)
    what = ak.Array([1.1, 2.2, 3.3], check_valid=True)

    assert to_list(ak.operations.with_field(recordarray, what)) == [
        (1, 1.1),
        (2, 2.2),
        (3, 3.3),
    ]
    assert to_list(ak.operations.with_field(recordarray, what, "a")) == [
        {"0": 1, "a": 1.1},
        {"0": 2, "a": 2.2},
        {"0": 3, "a": 3.3},
    ]
    assert ak.operations.fields(recordarray) == ["0"]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak
from awkward.types import ArrayType, ListType, NumpyType, OptionType, RegularType


def test_simple():
    a = ak.from_numpy(np.array([[1, 2], [3, 4], [5, 6]]), regulararray=True)
    b = ak.from_numpy(np.array([[7, 8], [9, 10]]), regulararray=True)
    c = a.layout._mergemany([b.layout])
    assert isinstance(c, ak.contents.RegularArray)
    assert c.size == 2
    assert ak.operations.to_list(c) == [[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]]


def test_regular_regular():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [[0.0, 1.1], [2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 2), 5)


def test_regular_option():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5], [6.6, 7.7], null, [8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [
        [0.0, 1.1],
        [2.2, 3.3],
        [4.4, 5.5],
        [6.6, 7.7],
        None,
        [8.8, 9.9],
    ]
    assert c.type == ArrayType(OptionType(RegularType(NumpyType("float64"), 2)), 6)


def test_option_regular():
    a1 = ak.from_json("[[0.0, 1.1], null, [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [
        [0.0, 1.1],
        None,
        [2.2, 3.3],
        [4.4, 5.5],
        [6.6, 7.7],
        [8.8, 9.9],
    ]
    assert c.type == ArrayType(OptionType(RegularType(NumpyType("float64"), 2)), 6)


def test_option_option():
    a1 = ak.from_json("[[0.0, 1.1], null, [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5], [6.6, 7.7], null, [8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [
        [0.0, 1.1],
        None,
        [2.2, 3.3],
        [4.4, 5.5],
        [6.6, 7.7],
        None,
        [8.8, 9.9],
    ]
    assert c.type == ArrayType(OptionType(RegularType(NumpyType("float64"), 2)), 7)


def test_regular_numpy():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.Array(np.array([[4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]))
    a1 = ak.to_regular(a1, axis=1)
    assert isinstance(a2.layout, ak.contents.NumpyArray)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [[0.0, 1.1], [2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 2), 5)


def test_numpy_regular():
    a1 = ak.Array(np.array([[0.0, 1.1], [2.2, 3.3]]))
    a2 = ak.from_json("[[4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]")
    assert isinstance(a1.layout, ak.contents.NumpyArray)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2])
    assert c.to_list() == [[0.0, 1.1], [2.2, 3.3], [4.4, 5.5], [6.6, 7.7], [8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 2), 5)


def test_regular_regular_axis1():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)


def test_option_regular_axis1():
    a1 = ak.from_json("[[0.0, 1.1], null, [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5, 6.6], [7, 8, 9], [7.7, 8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [
        [0.0, 1.1, 4.4, 5.5, 6.6],
        [7, 8, 9],
        [2.2, 3.3, 7.7, 8.8, 9.9],
    ]
    assert c.type == ArrayType(ListType(NumpyType("float64")), 3)


def test_regular_option_axis1():
    a1 = ak.from_json("[[0.0, 1.1], [7, 8], [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5, 6.6], null, [7.7, 8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [7, 8], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(ListType(NumpyType("float64")), 3)


def test_option_option_axis1():
    a1 = ak.from_json("[[0.0, 1.1], null, [2.2, 3.3]]")
    a2 = ak.from_json("[[4.4, 5.5, 6.6], null, [7.7, 8.8, 9.9]]")
    a1 = ak.to_regular(a1, axis=1)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(ListType(NumpyType("float64")), 3)


def test_regular_numpy_axis1():
    a1 = ak.from_json("[[0.0, 1.1], [2.2, 3.3]]")
    a2 = ak.Array(np.array([[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]))
    a1 = ak.to_regular(a1, axis=1)
    assert isinstance(a2.layout, ak.contents.NumpyArray)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)


def test_numpy_regular_axis1():
    a1 = ak.Array(np.array([[0.0, 1.1], [2.2, 3.3]]))
    a2 = ak.from_json("[[4.4, 5.5, 6.6], [7.7, 8.8, 9.9]]")
    assert isinstance(a1.layout, ak.contents.NumpyArray)
    a2 = ak.to_regular(a2, axis=1)
    c = ak.concatenate([a1, a2], axis=1)
    assert c.to_list() == [[0.0, 1.1, 4.4, 5.5, 6.6], [2.2, 3.3, 7.7, 8.8, 9.9]]
    assert c.type == ArrayType(RegularType(NumpyType("float64"), 5), 2)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test_no_parameter():
    dtype = np.dtype("datetime64[15us]")
    with pytest.warns(
        DeprecationWarning, match=r"from_dtype conversion of temporal units"
    ):
        form = ak.forms.numpyform.from_dtype(dtype)
    assert form.primitive == "datetime64"
    assert form.parameters == {"__unit__": "15us"}


def test_true_parameter():
    dtype = np.dtype("datetime64[15us]")
    with pytest.warns(
        DeprecationWarning, match=r"from_dtype conversion of temporal units"
    ):
        form = ak.forms.numpyform.from_dtype(dtype, time_units_as_parameter=True)
    assert form.primitive == "datetime64"
    assert form.parameters == {"__unit__": "15us"}


def test_false_parameter():
    dtype = np.dtype("datetime64[15us]")
    form = ak.forms.numpyform.from_dtype(dtype, time_units_as_parameter=False)
    assert form.primitive == "datetime64[15us]"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak


def test():
    assert ak.operations.is_none(ak.Array([1, 2, 3, None, 5])).to_list() == [
        False,
        False,
        False,
        True,
        False,
    ]
    assert ak.operations.is_none(ak.Array([[1, 2, 3], [], [None, 5]])).to_list() == [
        False,
        False,
        False,
    ]
    assert ak.operations.is_none(
        ak.Array([[1, 2, 3], [], [None, 5]]), axis=1
    ).to_list() == [
        [False, False, False],
        [],
        [True, False],
    ]
    assert ak.operations.is_none(
        ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=1
    ).to_list() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.operations.is_none(
        ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-1
    ).to_list() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.operations.is_none(
        ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-2
    ).to_list() == [
        False,
        False,
        False,
    ]
    assert ak.operations.is_none(
        ak.Array([[1, None, 2, 3], None, [None, 5]]), axis=-2
    ).to_list() == [False, True, False]
    with pytest.raises(ValueError):
        ak.operations.is_none(ak.Array([[1, None, 2, 3], [], [None, 5]]), axis=-3)

    one = ak.operations.from_iter([1, None, 3], highlevel=False)
    two = ak.operations.from_iter([[], [1], None, [3, 3, 3]], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 1, 0, 0, 1, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2, 3], dtype=np.int64))
    array = ak.Array(ak.contents.UnionArray(tags, index, [one, two]), check_valid=True)
    assert ak.to_list(array) == [1, [], [1], None, 3, None, [3, 3, 3]]

    assert ak.to_list(ak.operations.is_none(array)) == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

pyarrow = pytest.importorskip("pyarrow")

to_list = ak.operations.to_list


def test():
    import awkward._connect.pyarrow

    struct = pyarrow.struct(
        [
            pyarrow.field("x", pyarrow.list_(pyarrow.float64(), 2)),
            pyarrow.field("y", pyarrow.list_(pyarrow.float64(), 2)),
        ]
    )
    array = pyarrow.array(
        [
            {"x": [1.1, 2.1], "y": [3.1, 4.1]},
            {"x": [1.2, 2.2], "y": [3.2, 4.2]},
            {"x": [1.3, 2.3], "y": [3.3, 4.3]},
        ],
        type=struct,
    )
    as_awkward = awkward.from_arrow(array, highlevel=False)

    assert to_list(as_awkward) == [
        {"x": [1.1, 2.1], "y": [3.1, 4.1]},
        {"x": [1.2, 2.2], "y": [3.2, 4.2]},
        {"x": [1.3, 2.3], "y": [3.3, 4.3]},
    ]
    assert (
        str(as_awkward.form.type)
        == "{x: option[2 * ?float64], y: option[2 * ?float64]}"
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_to_RegularArray():
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 7, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 7 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 7, 11, 0), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 7 * 11 * 0 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 7, 0, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 7 * 0 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 0, 7, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 0 * 7 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 0, 5, 7, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 0 * 5 * 7 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 0, 3, 5, 7, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 0 * 3 * 5 * 7 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((0, 2, 3, 5, 7, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "0 * 2 * 3 * 5 * 7 * 11 * float64"
    )

    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 7, 0, 11, 0), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 7 * 0 * 11 * 0 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 5, 0, 7, 11, 0), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 5 * 0 * 7 * 11 * 0 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 3, 0, 5, 7, 0, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 3 * 0 * 5 * 7 * 0 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((2, 0, 3, 5, 7, 0, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "2 * 0 * 3 * 5 * 7 * 0 * 11 * float64"
    )
    assert (
        str(
            ak.operations.type(
                ak.highlevel.Array(
                    ak.operations.from_numpy(
                        np.empty((0, 2, 3, 5, 7, 0, 11), np.float64)
                    ).layout.to_RegularArray()
                )
            )
        )
        == "0 * 2 * 3 * 5 * 7 * 0 * 11 * float64"
    )


def test_actual():
    x = ak.operations.from_numpy(np.arange(2 * 3 * 4, dtype=np.int64).reshape(2, 3, 4))
    s = x[..., :0]
    result = ak.operations.zip({"q": s, "t": s})
    assert str(ak.operations.type(result)) == "2 * 3 * 0 * {q: int64, t: int64}"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

from typing import Optional

import numpy as np
import pytest

import awkward as ak


def assert_overrides_typestr(
    type_, typestr: str = "override", expected: Optional[str] = None
):
    # Assume typestr is expected result
    if expected is None:
        expected = typestr
    # Set appropriate array or record parameter
    if isinstance(type_, ak.types.RecordType):
        parameters = {**type_.parameters, "__record__": "custom"}
    else:
        assert isinstance(type_, (ak.types.ListType, ak.types.RegularType))
        parameters = {**type_.parameters, "__list__": "custom"}
    behavior = {("__typestr__", "custom"): typestr}
    # Build highlevel type with custom behavior
    with_parameter = type_.copy(parameters=parameters)
    as_str = "".join(with_parameter._str("", False, behavior))
    assert as_str == expected


def test_UnknownType():
    assert str(ak.types.unknowntype.UnknownType()) == "unknown"
    with pytest.raises(TypeError):
        ak.types.unknowntype.UnknownType(parameters={"x": 123})
    with pytest.raises(TypeError):
        ak.types.unknowntype.UnknownType(parameters={"__categorical__": True})
    assert repr(ak.types.unknowntype.UnknownType()) == "UnknownType()"


@pytest.mark.skipif(
    ak._util.win,
    reason="NumPy does not have float16, float128, and complex256 -- on Windows",
)
def test_NumpyType():
    assert str(ak.types.numpytype.NumpyType("bool")) == "bool"
    assert str(ak.types.numpytype.NumpyType("int8")) == "int8"
    assert str(ak.types.numpytype.NumpyType("uint8")) == "uint8"
    assert str(ak.types.numpytype.NumpyType("int16")) == "int16"
    assert str(ak.types.numpytype.NumpyType("uint16")) == "uint16"
    assert str(ak.types.numpytype.NumpyType("int32")) == "int32"
    assert str(ak.types.numpytype.NumpyType("uint32")) == "uint32"
    assert str(ak.types.numpytype.NumpyType("int64")) == "int64"
    assert str(ak.types.numpytype.NumpyType("uint64")) == "uint64"
    if hasattr(np, "float16"):
        assert str(ak.types.numpytype.NumpyType("float16")) == "float16"
    assert str(ak.types.numpytype.NumpyType("float32")) == "float32"
    assert str(ak.types.numpytype.NumpyType("float64")) == "float64"
    if hasattr(np, "float128"):
        assert str(ak.types.numpytype.NumpyType("float128")) == "float128"
    assert str(ak.types.numpytype.NumpyType("complex64")) == "complex64"
    assert str(ak.types.numpytype.NumpyType("complex128")) == "complex128"
    if hasattr(np, "complex256"):
        assert str(ak.types.numpytype.NumpyType("complex256")) == "complex256"
    assert (
        str(ak.types.numpytype.NumpyType("bool", parameters={"x": 123}))
        == 'bool[parameters={"x": 123}]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("bool", parameters={"__categorical__": True}))
        == "categorical[type=bool]"
    )
    assert (
        str(
            ak.types.numpytype.NumpyType(
                "bool", parameters={"__categorical__": True, "x": 123}
            )
        )
        == 'categorical[type=bool[parameters={"x": 123}]]'
    )

    assert str(ak.types.numpytype.NumpyType("datetime64")) == "datetime64"
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "Y"}))
        == 'datetime64[unit="Y"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "M"}))
        == 'datetime64[unit="M"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "W"}))
        == 'datetime64[unit="W"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "D"}))
        == 'datetime64[unit="D"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "h"}))
        == 'datetime64[unit="h"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "m"}))
        == 'datetime64[unit="m"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "s"}))
        == 'datetime64[unit="s"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "ms"}))
        == 'datetime64[unit="ms"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "us"}))
        == 'datetime64[unit="us"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "ns"}))
        == 'datetime64[unit="ns"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "ps"}))
        == 'datetime64[unit="ps"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "fs"}))
        == 'datetime64[unit="fs"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "as"}))
        == 'datetime64[unit="as"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "10s"}))
        == 'datetime64[unit="10s"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"__unit__": "1s"}))
        == 'datetime64[unit="s"]'
    )
    assert (
        str(
            ak.types.numpytype.NumpyType(
                "datetime64", parameters={"__unit__": "s", "x": 123}
            )
        )
        == 'datetime64[unit="s", parameters={"x": 123}]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("datetime64", parameters={"x": 123}))
        == 'datetime64[parameters={"x": 123}]'
    )
    assert str(ak.types.numpytype.NumpyType("timedelta64")) == "timedelta64"
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "Y"}))
        == 'timedelta64[unit="Y"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "M"}))
        == 'timedelta64[unit="M"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "W"}))
        == 'timedelta64[unit="W"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "D"}))
        == 'timedelta64[unit="D"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "h"}))
        == 'timedelta64[unit="h"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "m"}))
        == 'timedelta64[unit="m"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "s"}))
        == 'timedelta64[unit="s"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "ms"}))
        == 'timedelta64[unit="ms"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "us"}))
        == 'timedelta64[unit="us"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "ns"}))
        == 'timedelta64[unit="ns"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "ps"}))
        == 'timedelta64[unit="ps"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "fs"}))
        == 'timedelta64[unit="fs"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "as"}))
        == 'timedelta64[unit="as"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "10s"}))
        == 'timedelta64[unit="10s"]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"__unit__": "1s"}))
        == 'timedelta64[unit="s"]'
    )
    assert (
        str(
            ak.types.numpytype.NumpyType(
                "timedelta64", parameters={"__unit__": "s", "x": 123}
            )
        )
        == 'timedelta64[unit="s", parameters={"x": 123}]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("timedelta64", parameters={"x": 123}))
        == 'timedelta64[parameters={"x": 123}]'
    )
    assert (
        str(ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "char"}))
        == "char"
    )
    assert (
        str(ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "byte"}))
        == "byte"
    )
    assert repr(ak.types.numpytype.NumpyType(primitive="bool")) == "NumpyType('bool')"
    assert (
        repr(
            ak.types.numpytype.NumpyType(
                primitive="bool", parameters={"__categorical__": True}
            )
        )
        == "NumpyType('bool', parameters={'__categorical__': True})"
    )
    assert (
        repr(
            ak.types.numpytype.NumpyType(
                primitive="datetime64", parameters={"__unit__": "s"}
            )
        )
        == "NumpyType('datetime64', parameters={'__unit__': 's'})"
    )
    assert (
        repr(
            ak.types.numpytype.NumpyType(
                primitive="uint8", parameters={"__array__": "char"}
            )
        )
        == "NumpyType('uint8', parameters={'__array__': 'char'})"
    )
    assert (
        repr(
            ak.types.numpytype.NumpyType(
                primitive="uint8", parameters={"__array__": "byte"}
            )
        )
        == "NumpyType('uint8', parameters={'__array__': 'byte'})"
    )


def test_RegularType():
    assert (
        str(ak.types.regulartype.RegularType(ak.types.unknowntype.UnknownType(), 10))
        == "10 * unknown"
    )
    assert (
        str(ak.types.regulartype.RegularType(ak.types.unknowntype.UnknownType(), 0))
        == "0 * unknown"
    )
    with pytest.raises(ValueError):
        ak.types.regulartype.RegularType(ak.types.unknowntype.UnknownType(), -1)
    assert (
        str(
            ak.types.regulartype.RegularType(
                ak.types.unknowntype.UnknownType(), 10, parameters={"x": 123}
            )
        )
        == '[10 * unknown, parameters={"x": 123}]'
    )

    assert_overrides_typestr(
        ak.types.regulartype.RegularType(ak.types.unknowntype.UnknownType(), 10),
        expected="override[10]",
    )

    assert_overrides_typestr(
        ak.types.regulartype.RegularType(
            ak.types.unknowntype.UnknownType(),
            10,
            parameters={"x": 123},
        ),
        expected="override[10]",
    )
    assert (
        str(
            ak.types.regulartype.RegularType(
                ak.types.unknowntype.UnknownType(),
                10,
                parameters={"__categorical__": True},
            )
        )
        == "categorical[type=10 * unknown]"
    )
    assert (
        str(
            ak.types.regulartype.RegularType(
                ak.types.unknowntype.UnknownType(),
                10,
                parameters={"__categorical__": True, "x": 123},
            )
        )
        == 'categorical[type=[10 * unknown, parameters={"x": 123}]]'
    )
    assert_overrides_typestr(
        ak.types.regulartype.RegularType(
            ak.types.unknowntype.UnknownType(),
            10,
            parameters={"__categorical__": True},
        ),
        expected="categorical[type=override[10]]",
    )
    assert (
        str(
            ak.types.regulartype.RegularType(
                ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "char"}),
                10,
                parameters={"__array__": "string"},
            )
        )
        == "string[10]"
    )
    assert (
        str(
            ak.types.regulartype.RegularType(
                ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "byte"}),
                10,
                parameters={"__array__": "bytestring"},
            )
        )
        == "bytes[10]"
    )

    assert (
        repr(
            ak.types.regulartype.RegularType(
                content=ak.types.unknowntype.UnknownType(), size=10
            )
        )
        == "RegularType(UnknownType(), 10)"
    )
    assert (
        repr(
            ak.types.regulartype.RegularType(
                content=ak.types.unknowntype.UnknownType(),
                size=10,
                parameters={"__categorical__": True},
            )
        )
        == "RegularType(UnknownType(), 10, parameters={'__categorical__': True})"
    )
    assert (
        repr(
            ak.types.regulartype.RegularType(
                content=ak.types.numpytype.NumpyType(
                    primitive="uint8", parameters={"__array__": "char"}
                ),
                parameters={"__array__": "string"},
                size=10,
            )
        )
        == "RegularType(NumpyType('uint8', parameters={'__array__': 'char'}), 10, parameters={'__array__': 'string'})"
    )
    assert (
        repr(
            ak.types.regulartype.RegularType(
                content=ak.types.numpytype.NumpyType(
                    primitive="uint8", parameters={"__array__": "byte"}
                ),
                parameters={"__array__": "bytestring"},
                size=10,
            )
        )
        == "RegularType(NumpyType('uint8', parameters={'__array__': 'byte'}), 10, parameters={'__array__': 'bytestring'})"
    )


def test_ListType():
    assert (
        str(ak.types.listtype.ListType(ak.types.unknowntype.UnknownType()))
        == "var * unknown"
    )
    assert (
        str(
            ak.types.listtype.ListType(
                ak.types.unknowntype.UnknownType(), parameters={"x": 123}
            )
        )
        == '[var * unknown, parameters={"x": 123}]'
    )
    assert_overrides_typestr(
        ak.types.listtype.ListType(ak.types.unknowntype.UnknownType())
    )
    assert_overrides_typestr(
        ak.types.listtype.ListType(
            ak.types.unknowntype.UnknownType(),
            parameters={"x": 123},
        )
    )
    assert (
        str(
            ak.types.listtype.ListType(
                ak.types.unknowntype.UnknownType(), parameters={"__categorical__": True}
            )
        )
        == "categorical[type=var * unknown]"
    )
    assert (
        str(
            ak.types.listtype.ListType(
                ak.types.unknowntype.UnknownType(),
                parameters={"__categorical__": True, "x": 123},
            )
        )
        == 'categorical[type=[var * unknown, parameters={"x": 123}]]'
    )
    assert_overrides_typestr(
        ak.types.listtype.ListType(
            ak.types.unknowntype.UnknownType(),
            parameters={"__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert (
        str(
            ak.types.listtype.ListType(
                ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "char"}),
                parameters={"__array__": "string"},
            )
        )
        == "string"
    )
    assert (
        str(
            ak.types.listtype.ListType(
                ak.types.numpytype.NumpyType("uint8", parameters={"__array__": "byte"}),
                parameters={"__array__": "bytestring"},
            )
        )
        == "bytes"
    )

    assert (
        repr(ak.types.listtype.ListType(content=ak.types.unknowntype.UnknownType()))
        == "ListType(UnknownType())"
    )
    assert (
        repr(
            ak.types.listtype.ListType(
                content=ak.types.unknowntype.UnknownType(),
                parameters={"__categorical__": True},
            )
        )
        == "ListType(UnknownType(), parameters={'__categorical__': True})"
    )
    assert (
        repr(
            ak.types.listtype.ListType(
                content=ak.types.numpytype.NumpyType(
                    primitive="uint8", parameters={"__array__": "char"}
                ),
                parameters={"__array__": "string"},
            )
        )
        == "ListType(NumpyType('uint8', parameters={'__array__': 'char'}), parameters={'__array__': 'string'})"
    )
    assert (
        repr(
            ak.types.listtype.ListType(
                content=ak.types.numpytype.NumpyType(
                    primitive="uint8", parameters={"__array__": "byte"}
                ),
                parameters={"__array__": "bytestring"},
            )
        )
        == "ListType(NumpyType('uint8', parameters={'__array__': 'byte'}), parameters={'__array__': 'bytestring'})"
    )


def test_RecordType():
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
            )
        )
        == "(unknown, bool)"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
            )
        )
        == "{x: unknown, y: bool}"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"__record__": "Name"},
            )
        )
        == "Name[unknown, bool]"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"__record__": "Name"},
            )
        )
        == "Name[x: unknown, y: bool]"
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"__record__": "Name"},
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"__record__": "Name"},
        )
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"x": 123},
            )
        )
        == 'tuple[[unknown, bool], parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"x": 123},
            )
        )
        == 'struct[{x: unknown, y: bool}, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"__record__": "Name", "x": 123},
            )
        )
        == 'Name[unknown, bool, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"__record__": "Name", "x": 123},
            )
        )
        == 'Name[x: unknown, y: bool, parameters={"x": 123}]'
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"x": 123},
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"x": 123},
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"__record__": "Name", "x": 123},
        )
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"__record__": "Name", "x": 123},
        )
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"__categorical__": True},
            )
        )
        == "categorical[type=(unknown, bool)]"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"__categorical__": True},
            )
        )
        == "categorical[type={x: unknown, y: bool}]"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"__record__": "Name", "__categorical__": True},
            )
        )
        == "categorical[type=Name[unknown, bool]]"
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"__record__": "Name", "__categorical__": True},
            )
        )
        == "categorical[type=Name[x: unknown, y: bool]]"
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"__record__": "Name", "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"__record__": "Name", "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'categorical[type=tuple[[unknown, bool], parameters={"x": 123}]]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'categorical[type=struct[{x: unknown, y: bool}, parameters={"x": 123}]]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                None,
                parameters={"__record__": "Name", "x": 123, "__categorical__": True},
            )
        )
        == 'categorical[type=Name[unknown, bool, parameters={"x": 123}]]'
    )
    assert (
        str(
            ak.types.recordtype.RecordType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                ["x", "y"],
                parameters={"__record__": "Name", "x": 123, "__categorical__": True},
            )
        )
        == 'categorical[type=Name[x: unknown, y: bool, parameters={"x": 123}]]'
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"x": 123, "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"x": 123, "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            None,
            parameters={"__record__": "Name", "x": 123, "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )
    assert_overrides_typestr(
        ak.types.recordtype.RecordType(
            [
                ak.types.unknowntype.UnknownType(),
                ak.types.numpytype.NumpyType("bool"),
            ],
            ["x", "y"],
            parameters={"__record__": "Name", "x": 123, "__categorical__": True},
        ),
        expected="categorical[type=override]",
    )

    assert (
        repr(
            ak.types.recordtype.RecordType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                fields=None,
            )
        )
        == "RecordType([UnknownType(), NumpyType('bool')], None)"
    )
    assert (
        repr(
            ak.types.recordtype.RecordType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                fields=["x", "y"],
            )
        )
        == "RecordType([UnknownType(), NumpyType('bool')], ['x', 'y'])"
    )
    assert (
        repr(
            ak.types.recordtype.RecordType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                fields=None,
                parameters={"__record__": "Name", "x": 123, "__categorical__": True},
            )
        )
        == "RecordType([UnknownType(), NumpyType('bool')], None, parameters={'__record__': 'Name', 'x': 123, '__categorical__': True})"
    )
    assert (
        repr(
            ak.types.recordtype.RecordType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                fields=None,
                parameters={"__record__": "Name", "x": 123, "__categorical__": True},
            )
        )
        == "RecordType([UnknownType(), NumpyType('bool')], None, parameters={'__record__': 'Name', 'x': 123, '__categorical__': True})"
    )
    assert (
        repr(
            ak.types.recordtype.RecordType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                fields=["x", "y"],
                parameters={"__record__": "Name", "x": 123, "__categorical__": True},
            )
        )
        == "RecordType([UnknownType(), NumpyType('bool')], ['x', 'y'], parameters={'__record__': 'Name', 'x': 123, '__categorical__': True})"
    )


def test_OptionType():
    assert (
        str(ak.types.optiontype.OptionType(ak.types.unknowntype.UnknownType()))
        == "?unknown"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.listtype.ListType(ak.types.unknowntype.UnknownType())
            )
        )
        == "option[var * unknown]"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.regulartype.RegularType(ak.types.unknowntype.UnknownType(), 10)
            )
        )
        == "option[10 * unknown]"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.unknowntype.UnknownType(), parameters={"x": 123}
            )
        )
        == 'option[unknown, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.listtype.ListType(ak.types.unknowntype.UnknownType()),
                parameters={"x": 123},
            )
        )
        == 'option[var * unknown, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.regulartype.RegularType(
                    ak.types.unknowntype.UnknownType(), 10
                ),
                parameters={"x": 123},
            )
        )
        == 'option[10 * unknown, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.unknowntype.UnknownType(), parameters={"__categorical__": True}
            )
        )
        == "?categorical[type=unknown]"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.listtype.ListType(ak.types.unknowntype.UnknownType()),
                parameters={"__categorical__": True},
            )
        )
        == "option[categorical[type=var * unknown]]"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.regulartype.RegularType(
                    ak.types.unknowntype.UnknownType(), 10
                ),
                parameters={"__categorical__": True},
            )
        )
        == "option[categorical[type=10 * unknown]]"
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.unknowntype.UnknownType(),
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'option[categorical[type=unknown], parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.listtype.ListType(ak.types.unknowntype.UnknownType()),
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'option[categorical[type=var * unknown], parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.optiontype.OptionType(
                ak.types.regulartype.RegularType(
                    ak.types.unknowntype.UnknownType(), 10
                ),
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'option[categorical[type=10 * unknown], parameters={"x": 123}]'
    )

    assert (
        repr(ak.types.optiontype.OptionType(content=ak.types.unknowntype.UnknownType()))
        == "OptionType(UnknownType())"
    )
    assert (
        repr(
            ak.types.optiontype.OptionType(
                content=ak.types.listtype.ListType(ak.types.unknowntype.UnknownType())
            )
        )
        == "OptionType(ListType(UnknownType()))"
    )
    assert (
        repr(
            ak.types.optiontype.OptionType(
                content=ak.types.regulartype.RegularType(
                    ak.types.unknowntype.UnknownType(), 10
                ),
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == "OptionType(RegularType(UnknownType(), 10), parameters={'x': 123, '__categorical__': True})"
    )


def test_UnionType():
    assert (
        str(
            ak.types.uniontype.UnionType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ]
            )
        )
        == "union[unknown, bool]"
    )
    assert (
        str(
            ak.types.uniontype.UnionType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                parameters={"x": 123},
            )
        )
        == 'union[unknown, bool, parameters={"x": 123}]'
    )
    assert (
        str(
            ak.types.uniontype.UnionType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                parameters={"__categorical__": True},
            )
        )
        == "categorical[type=union[unknown, bool]]"
    )
    assert (
        str(
            ak.types.uniontype.UnionType(
                [
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == 'categorical[type=union[unknown, bool, parameters={"x": 123}]]'
    )

    assert (
        repr(
            ak.types.uniontype.UnionType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ]
            )
        )
        == "UnionType([UnknownType(), NumpyType('bool')])"
    )
    assert (
        repr(
            ak.types.uniontype.UnionType(
                contents=[
                    ak.types.unknowntype.UnknownType(),
                    ak.types.numpytype.NumpyType("bool"),
                ],
                parameters={"x": 123, "__categorical__": True},
            )
        )
        == "UnionType([UnknownType(), NumpyType('bool')], parameters={'x': 123, '__categorical__': True})"
    )


def test_ArrayType():
    assert (
        str(ak.types.arraytype.ArrayType(ak.types.unknowntype.UnknownType(), 10))
        == "10 * unknown"
    )
    assert (
        str(ak.types.arraytype.ArrayType(ak.types.unknowntype.UnknownType(), 0))
        == "0 * unknown"
    )
    with pytest.raises(ValueError):
        ak.types.arraytype.ArrayType(ak.types.unknowntype.UnknownType(), -1)

    # ArrayType should not have these arguments (should not be a Type subclass)
    with pytest.raises(TypeError):
        ak.types.arraytype.ArrayType(
            ak.types.arraytype.ArrayType(ak.types.unknowntype.UnknownType(), 10), 10
        )

    assert (
        repr(
            ak.types.arraytype.ArrayType(
                content=ak.types.unknowntype.UnknownType(), length=10
            )
        )
        == "ArrayType(UnknownType(), 10, None)"
    )

    assert (
        str(
            ak.types.arraytype.ArrayType(
                content=ak.types.numpytype.NumpyType(
                    "int64", parameters={"catastrophe": "apocalypse"}
                ),
                length=10,
            )
        )
        == '10 * int64[parameters={"catastrophe": "apocalypse"}]'
    )


def test_EmptyForm():
    assert (
        str(ak.forms.emptyform.EmptyForm())
        == """{
    "class": "EmptyArray"
}"""
    )
    assert (
        str(ak.forms.emptyform.EmptyForm(form_key="hello"))
        == """{
    "class": "EmptyArray",
    "form_key": "hello"
}"""
    )
    assert repr(ak.forms.emptyform.EmptyForm()) == "EmptyForm()"
    with pytest.raises(TypeError):
        ak.forms.emptyform.EmptyForm(parameters={"x": 123}, form_key="hello")

    assert ak.forms.emptyform.EmptyForm().to_dict(verbose=False) == {
        "class": "EmptyArray"
    }
    assert ak.forms.emptyform.EmptyForm().to_dict() == {
        "class": "EmptyArray",
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.emptyform.EmptyForm(form_key="hello").to_dict(verbose=False) == {
        "class": "EmptyArray",
        "form_key": "hello",
    }
    assert ak.forms.from_dict({"class": "EmptyArray"}).to_dict() == {
        "class": "EmptyArray",
        "parameters": {},
        "form_key": None,
    }


@pytest.mark.skipif(
    ak._util.win,
    reason="NumPy does not have float16, float128, and complex256 -- on Windows",
)
def test_NumpyForm():
    assert (
        str(ak.forms.numpyform.NumpyForm("bool"))
        == """{
    "class": "NumpyArray",
    "primitive": "bool"
}"""
    )
    assert repr(ak.forms.numpyform.NumpyForm(primitive="bool")) == "NumpyForm('bool')"
    assert (
        repr(
            ak.forms.numpyform.NumpyForm(
                primitive="bool",
                inner_shape=[1, 2, 3],
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "NumpyForm('bool', inner_shape=(1, 2, 3), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.numpyform.NumpyForm("bool").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "bool",
    }
    assert ak.forms.numpyform.NumpyForm("bool").to_dict() == {
        "class": "NumpyArray",
        "primitive": "bool",
        "inner_shape": [],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.numpyform.NumpyForm(
        "bool",
        inner_shape=[1, 2, 3],
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "bool",
        "inner_shape": [1, 2, 3],
        "parameters": {"x": 123},
        "form_key": "hello",
    }

    assert ak.forms.numpyform.NumpyForm("bool").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "bool",
    }
    assert ak.forms.numpyform.NumpyForm("int8").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int8",
    }
    assert ak.forms.numpyform.NumpyForm("uint8").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint8",
    }
    assert ak.forms.numpyform.NumpyForm("int16").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int16",
    }
    assert ak.forms.numpyform.NumpyForm("uint16").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint16",
    }
    assert ak.forms.numpyform.NumpyForm("int32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int32",
    }
    assert ak.forms.numpyform.NumpyForm("uint32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint32",
    }
    assert ak.forms.numpyform.NumpyForm("int64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int64",
    }
    assert ak.forms.numpyform.NumpyForm("uint64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint64",
    }
    if hasattr(np, "float16"):
        assert ak.forms.numpyform.NumpyForm("float16").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float16",
        }
    assert ak.forms.numpyform.NumpyForm("float32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "float32",
    }
    assert ak.forms.numpyform.NumpyForm("float64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "float64",
    }
    if hasattr(np, "float128"):
        assert ak.forms.numpyform.NumpyForm("float128").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float128",
        }
    assert ak.forms.numpyform.NumpyForm("complex64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex64",
    }
    assert ak.forms.numpyform.NumpyForm("complex128").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex128",
    }
    if hasattr(np, "complex256"):
        assert ak.forms.numpyform.NumpyForm("complex256").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "complex256",
        }
    assert ak.forms.numpyform.NumpyForm("datetime64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
    }
    assert ak.forms.numpyform.NumpyForm(
        "datetime64", parameters={"__unit__": "s"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
        "parameters": {"__unit__": "s"},
    }
    assert ak.forms.numpyform.NumpyForm(
        "datetime64", parameters={"__unit__": "s", "x": 123}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
        "parameters": {"__unit__": "s", "x": 123},
    }
    assert ak.forms.numpyform.NumpyForm("timedelta64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
    }
    assert ak.forms.numpyform.NumpyForm(
        "timedelta64", parameters={"__unit__": "s"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
        "parameters": {"__unit__": "s"},
    }
    assert ak.forms.numpyform.NumpyForm(
        "timedelta64", parameters={"__unit__": "s", "x": 123}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
        "parameters": {"__unit__": "s", "x": 123},
    }

    with pytest.warns(DeprecationWarning):
        assert ak.forms.numpyform.from_dtype(np.dtype("bool")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "bool",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("int8")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "int8",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("uint8")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "uint8",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("int16")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "int16",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("uint16")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "uint16",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("int32")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "int32",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("uint32")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "uint32",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("int64")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "int64",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("uint64")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "uint64",
        }
        if hasattr(np, "float16"):
            assert ak.forms.numpyform.from_dtype(np.dtype("float16")).to_dict(
                verbose=False
            ) == {
                "class": "NumpyArray",
                "primitive": "float16",
            }
        assert ak.forms.numpyform.from_dtype(np.dtype("float32")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "float32",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("float64")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "float64",
        }
        if hasattr(np, "float128"):
            assert ak.forms.numpyform.from_dtype(np.dtype("float128")).to_dict(
                verbose=False
            ) == {
                "class": "NumpyArray",
                "primitive": "float128",
            }
        assert ak.forms.numpyform.from_dtype(np.dtype("complex64")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "complex64",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("complex128")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "complex128",
        }
        if hasattr(np, "complex256"):
            assert ak.forms.numpyform.from_dtype(np.dtype("complex256")).to_dict(
                verbose=False
            ) == {
                "class": "NumpyArray",
                "primitive": "complex256",
            }
        assert ak.forms.numpyform.from_dtype(np.dtype("M8")).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "datetime64",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("M8[s]")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "datetime64",
            "parameters": {"__unit__": "s"},
        }
        assert ak.forms.numpyform.from_dtype(
            np.dtype("M8[s]"), parameters={"x": 123}
        ).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "datetime64",
            "parameters": {"__unit__": "s", "x": 123},
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("m8")).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "timedelta64",
        }
        assert ak.forms.numpyform.from_dtype(np.dtype("m8[s]")).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "timedelta64",
            "parameters": {"__unit__": "s"},
        }
        assert ak.forms.numpyform.from_dtype(
            np.dtype("m8[s]"), parameters={"x": 123}
        ).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "timedelta64",
            "parameters": {"__unit__": "s", "x": 123},
        }
        assert ak.forms.numpyform.from_dtype(np.dtype(("bool", (1, 2, 3)))).to_dict(
            verbose=False
        ) == {
            "class": "NumpyArray",
            "primitive": "bool",
            "inner_shape": [1, 2, 3],
        }

    assert ak.forms.numpyform.from_dtype(
        np.dtype("M8"), time_units_as_parameter=False
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
    }
    assert ak.forms.numpyform.from_dtype(
        np.dtype("M8[s]"), time_units_as_parameter=False
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64[s]",
    }
    assert ak.forms.numpyform.from_dtype(
        np.dtype("M8[s]"), time_units_as_parameter=False, parameters={"x": 123}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64[s]",
        "parameters": {"x": 123},
    }
    assert ak.forms.numpyform.from_dtype(
        np.dtype("m8"), time_units_as_parameter=False
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
    }
    assert ak.forms.numpyform.from_dtype(
        np.dtype("m8[s]"), time_units_as_parameter=False
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64[s]",
    }
    assert ak.forms.numpyform.from_dtype(
        np.dtype("m8[s]"), parameters={"x": 123}, time_units_as_parameter=False
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64[s]",
        "parameters": {"x": 123},
    }
    with pytest.raises(TypeError):
        ak.forms.from_dtype(np.dtype("O"), time_units_as_parameter=False).to_dict(
            verbose=False
        )
    with pytest.raises(TypeError):
        ak.forms.from_dtype(
            np.dtype([("one", np.int64), ("two", np.float64)]),
            time_units_as_parameter=False,
        ).to_dict(verbose=False)
    assert ak.forms.from_dict("bool").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "bool",
    }
    assert ak.forms.from_dict("int8").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int8",
    }
    assert ak.forms.from_dict("uint8").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint8",
    }
    assert ak.forms.from_dict("int16").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int16",
    }
    assert ak.forms.from_dict("uint16").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint16",
    }
    assert ak.forms.from_dict("int32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int32",
    }
    assert ak.forms.from_dict("uint32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint32",
    }
    assert ak.forms.from_dict("int64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "int64",
    }
    assert ak.forms.from_dict("uint64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "uint64",
    }
    if hasattr(np, "float16"):
        assert ak.forms.from_dict("float16").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float16",
        }
    assert ak.forms.from_dict("float32").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "float32",
    }
    assert ak.forms.from_dict("float64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "float64",
    }
    if hasattr(np, "float128"):
        assert ak.forms.from_dict("float128").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float128",
        }
    assert ak.forms.from_dict("complex64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex64",
    }
    assert ak.forms.from_dict("complex128").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex128",
    }
    if hasattr(np, "complex256"):
        assert ak.forms.from_dict("complex256").to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "complex256",
        }
    assert ak.forms.from_dict("datetime64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
    }
    assert ak.forms.from_dict(
        {
            "class": "NumpyArray",
            "primitive": "datetime64",
            "parameters": {"__unit__": "s"},
        }
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
        "parameters": {"__unit__": "s"},
    }
    assert ak.forms.from_dict(
        {
            "class": "NumpyArray",
            "primitive": "datetime64",
            "parameters": {"__unit__": "s", "x": 123},
        }
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
        "parameters": {"__unit__": "s", "x": 123},
    }
    assert ak.forms.from_dict("timedelta64").to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
    }
    assert ak.forms.from_dict(
        {
            "class": "NumpyArray",
            "primitive": "timedelta64",
            "parameters": {"__unit__": "s"},
        }
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
        "parameters": {"__unit__": "s"},
    }
    assert ak.forms.from_dict(
        {
            "class": "NumpyArray",
            "primitive": "timedelta64",
            "parameters": {"__unit__": "s", "x": 123},
        }
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
        "parameters": {"__unit__": "s", "x": 123},
    }

    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "bool"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "bool",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "int8"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "int8",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "uint8"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "uint8",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "int16"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "int16",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "uint16"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "uint16",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "int32"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "int32",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "uint32"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "uint32",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "int64"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "int64",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "uint64"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "uint64",
    }
    if hasattr(np, "float16"):
        assert ak.forms.from_dict(
            {"class": "NumpyArray", "primitive": "float16"}
        ).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float16",
        }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "float32"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "float32",
    }
    assert ak.forms.from_dict({"class": "NumpyArray", "primitive": "float64"}).to_dict(
        verbose=False
    ) == {
        "class": "NumpyArray",
        "primitive": "float64",
    }
    if hasattr(np, "float128"):
        assert ak.forms.from_dict(
            {"class": "NumpyArray", "primitive": "float128"}
        ).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "float128",
        }
    assert ak.forms.from_dict(
        {"class": "NumpyArray", "primitive": "complex64"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex64",
    }
    assert ak.forms.from_dict(
        {"class": "NumpyArray", "primitive": "complex128"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "complex128",
    }
    if hasattr(np, "complex256"):
        assert ak.forms.from_dict(
            {"class": "NumpyArray", "primitive": "complex256"}
        ).to_dict(verbose=False) == {
            "class": "NumpyArray",
            "primitive": "complex256",
        }
    assert ak.forms.from_dict(
        {"class": "NumpyArray", "primitive": "datetime64"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "datetime64",
    }
    assert ak.forms.from_dict(
        {"class": "NumpyArray", "primitive": "timedelta64"}
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "timedelta64",
    }

    assert ak.forms.from_dict(
        {
            "class": "NumpyArray",
            "primitive": "bool",
            "inner_shape": [1, 2, 3],
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "NumpyArray",
        "primitive": "bool",
        "inner_shape": [1, 2, 3],
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_RegularForm():
    assert (
        str(ak.forms.regularform.RegularForm(ak.forms.emptyform.EmptyForm(), 10))
        == """{
    "class": "RegularArray",
    "size": 10,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.regularform.RegularForm(
                ak.forms.emptyform.EmptyForm(),
                10,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "RegularArray",
    "size": 10,
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.regularform.RegularForm(
                content=ak.forms.emptyform.EmptyForm(), size=10
            )
        )
        == "RegularForm(EmptyForm(), 10)"
    )
    assert (
        repr(
            ak.forms.regularform.RegularForm(
                content=ak.forms.emptyform.EmptyForm(),
                size=10,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "RegularForm(EmptyForm(), 10, parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.regularform.RegularForm(ak.forms.emptyform.EmptyForm(), 10).to_dict(
        verbose=False
    ) == {
        "class": "RegularArray",
        "size": 10,
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.regularform.RegularForm(
        ak.forms.emptyform.EmptyForm(), 10
    ).to_dict() == {
        "class": "RegularArray",
        "size": 10,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.regularform.RegularForm(
        content=ak.forms.emptyform.EmptyForm(),
        size=10,
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "RegularArray",
        "size": 10,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {"class": "RegularArray", "size": 10, "content": {"class": "EmptyArray"}}
    ).to_dict() == {
        "class": "RegularArray",
        "size": 10,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "RegularArray",
            "size": 10,
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "RegularArray",
        "size": 10,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.regularform.RegularForm(
        ak.forms.numpyform.NumpyForm("bool"), 10
    ).to_dict() == {
        "class": "RegularArray",
        "content": {
            "class": "NumpyArray",
            "primitive": "bool",
            "inner_shape": [],
            "parameters": {},
            "form_key": None,
        },
        "size": 10,
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.regularform.RegularForm(
        ak.forms.numpyform.NumpyForm("bool"), 10
    ).to_dict(verbose=False) == {
        "class": "RegularArray",
        "content": "bool",
        "size": 10,
    }


def test_ListForm():
    assert (
        str(ak.forms.listform.ListForm("i32", "i32", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "ListArray",
    "starts": "i32",
    "stops": "i32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(ak.forms.listform.ListForm("u32", "u32", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "ListArray",
    "starts": "u32",
    "stops": "u32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(ak.forms.listform.ListForm("i64", "i64", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "ListArray",
    "starts": "i64",
    "stops": "i64",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.listform.ListForm(
                "i32",
                "i32",
                ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "ListArray",
    "starts": "i32",
    "stops": "i32",
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.listform.ListForm(
                starts="i32", stops="i32", content=ak.forms.emptyform.EmptyForm()
            )
        )
        == "ListForm('i32', 'i32', EmptyForm())"
    )
    assert (
        repr(
            ak.forms.listform.ListForm(
                starts="i32",
                stops="i32",
                content=ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "ListForm('i32', 'i32', EmptyForm(), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.listform.ListForm(
        "i32", "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict(verbose=False) == {
        "class": "ListArray",
        "starts": "i32",
        "stops": "i32",
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.listform.ListForm(
        "i32", "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict() == {
        "class": "ListArray",
        "starts": "i32",
        "stops": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.listform.ListForm(
        starts="i32",
        stops="i32",
        content=ak.forms.emptyform.EmptyForm(),
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "ListArray",
        "starts": "i32",
        "stops": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "ListArray",
            "starts": "i32",
            "stops": "i32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListArray",
        "starts": "i32",
        "stops": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListArray",
            "starts": "u32",
            "stops": "u32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListArray",
        "starts": "u32",
        "stops": "u32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListArray",
            "starts": "i64",
            "stops": "i64",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListArray",
        "starts": "i64",
        "stops": "i64",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListArray",
            "starts": "i32",
            "stops": "i32",
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "ListArray",
        "starts": "i32",
        "stops": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_ListOffsetForm():
    assert (
        str(
            ak.forms.listoffsetform.ListOffsetForm(
                "i32", ak.forms.emptyform.EmptyForm()
            )
        )
        == """{
    "class": "ListOffsetArray",
    "offsets": "i32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.listoffsetform.ListOffsetForm(
                "u32", ak.forms.emptyform.EmptyForm()
            )
        )
        == """{
    "class": "ListOffsetArray",
    "offsets": "u32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.listoffsetform.ListOffsetForm(
                "i64", ak.forms.emptyform.EmptyForm()
            )
        )
        == """{
    "class": "ListOffsetArray",
    "offsets": "i64",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.listoffsetform.ListOffsetForm(
                "i32",
                ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "ListOffsetArray",
    "offsets": "i32",
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.listoffsetform.ListOffsetForm(
                offsets="i32", content=ak.forms.emptyform.EmptyForm()
            )
        )
        == "ListOffsetForm('i32', EmptyForm())"
    )
    assert (
        repr(
            ak.forms.listoffsetform.ListOffsetForm(
                offsets="i32",
                content=ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "ListOffsetForm('i32', EmptyForm(), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.listoffsetform.ListOffsetForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict(verbose=False) == {
        "class": "ListOffsetArray",
        "offsets": "i32",
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.listoffsetform.ListOffsetForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict() == {
        "class": "ListOffsetArray",
        "offsets": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.listoffsetform.ListOffsetForm(
        offsets="i32",
        content=ak.forms.emptyform.EmptyForm(),
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "ListOffsetArray",
        "offsets": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "ListOffsetArray",
            "offsets": "i32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListOffsetArray",
        "offsets": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListOffsetArray",
            "offsets": "u32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListOffsetArray",
        "offsets": "u32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListOffsetArray",
            "offsets": "i64",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ListOffsetArray",
        "offsets": "i64",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ListOffsetArray",
            "offsets": "i32",
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "ListOffsetArray",
        "offsets": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_RecordForm():
    assert (
        str(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                None,
            )
        )
        == """{
    "class": "RecordArray",
    "fields": null,
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ]
}"""
    )
    assert (
        str(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                ["x", "y"],
            )
        )
        == """{
    "class": "RecordArray",
    "fields": [
        "x",
        "y"
    ],
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ]
}"""
    )
    assert (
        str(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                None,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "RecordArray",
    "fields": null,
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ],
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        str(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                ["x", "y"],
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "RecordArray",
    "fields": [
        "x",
        "y"
    ],
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ],
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )

    assert (
        repr(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                None,
            )
        )
        == "RecordForm([EmptyForm(), NumpyForm('bool')], None)"
    )
    assert (
        repr(
            ak.forms.recordform.RecordForm(
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                ["x", "y"],
            )
        )
        == "RecordForm([EmptyForm(), NumpyForm('bool')], ['x', 'y'])"
    )
    assert (
        repr(
            ak.forms.recordform.RecordForm(
                contents=[
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                fields=None,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "RecordForm([EmptyForm(), NumpyForm('bool')], None, parameters={'x': 123}, form_key='hello')"
    )
    assert (
        repr(
            ak.forms.recordform.RecordForm(
                contents=[
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                fields=["x", "y"],
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "RecordForm([EmptyForm(), NumpyForm('bool')], ['x', 'y'], parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.recordform.RecordForm(
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
        None,
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": None,
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
    }
    assert ak.forms.recordform.RecordForm(
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
        ["x", "y"],
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": ["x", "y"],
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
    }
    assert ak.forms.recordform.RecordForm(
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
        None,
    ).to_dict() == {
        "class": "RecordArray",
        "fields": None,
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.recordform.RecordForm(
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
        ["x", "y"],
    ).to_dict() == {
        "class": "RecordArray",
        "fields": ["x", "y"],
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.recordform.RecordForm(
        contents=[
            ak.forms.emptyform.EmptyForm(),
            ak.forms.numpyform.NumpyForm("bool"),
        ],
        fields=None,
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": None,
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.recordform.RecordForm(
        contents=[
            ak.forms.emptyform.EmptyForm(),
            ak.forms.numpyform.NumpyForm("bool"),
        ],
        fields=["x", "y"],
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": ["x", "y"],
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": None,
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
        }
    ).to_dict() == {
        "class": "RecordArray",
        "fields": None,
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": ["x", "y"],
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
        }
    ).to_dict() == {
        "class": "RecordArray",
        "fields": ["x", "y"],
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": None,
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": None,
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "RecordArray",
            "fields": ["x", "y"],
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "RecordArray",
        "fields": ["x", "y"],
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_IndexedForm():
    assert (
        str(ak.forms.indexedform.IndexedForm("i32", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "IndexedArray",
    "index": "i32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(ak.forms.indexedform.IndexedForm("u32", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "IndexedArray",
    "index": "u32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(ak.forms.indexedform.IndexedForm("i64", ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "IndexedArray",
    "index": "i64",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.indexedform.IndexedForm(
                "i32",
                ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "IndexedArray",
    "index": "i32",
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.indexedform.IndexedForm(
                index="i32", content=ak.forms.emptyform.EmptyForm()
            )
        )
        == "IndexedForm('i32', EmptyForm())"
    )
    assert (
        repr(
            ak.forms.indexedform.IndexedForm(
                index="i32",
                content=ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "IndexedForm('i32', EmptyForm(), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.indexedform.IndexedForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict(verbose=False) == {
        "class": "IndexedArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.indexedform.IndexedForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict() == {
        "class": "IndexedArray",
        "index": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.indexedform.IndexedForm(
        index="i32",
        content=ak.forms.emptyform.EmptyForm(),
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "IndexedArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedArray",
            "index": "i32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "IndexedArray",
        "index": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedArray",
            "index": "u32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "IndexedArray",
        "index": "u32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedArray",
            "index": "i64",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "IndexedArray",
        "index": "i64",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedArray",
            "index": "i32",
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "IndexedArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_IndexedOptionForm():
    assert (
        str(
            ak.forms.indexedoptionform.IndexedOptionForm(
                "i32", ak.forms.emptyform.EmptyForm()
            )
        )
        == """{
    "class": "IndexedOptionArray",
    "index": "i32",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.indexedoptionform.IndexedOptionForm(
                "i64", ak.forms.emptyform.EmptyForm()
            )
        )
        == """{
    "class": "IndexedOptionArray",
    "index": "i64",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.indexedoptionform.IndexedOptionForm(
                "i32",
                ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "IndexedOptionArray",
    "index": "i32",
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.indexedoptionform.IndexedOptionForm(
                index="i32", content=ak.forms.emptyform.EmptyForm()
            )
        )
        == "IndexedOptionForm('i32', EmptyForm())"
    )
    assert (
        repr(
            ak.forms.indexedoptionform.IndexedOptionForm(
                index="i32",
                content=ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "IndexedOptionForm('i32', EmptyForm(), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.indexedoptionform.IndexedOptionForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict(verbose=False) == {
        "class": "IndexedOptionArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.indexedoptionform.IndexedOptionForm(
        "i32", ak.forms.emptyform.EmptyForm()
    ).to_dict() == {
        "class": "IndexedOptionArray",
        "index": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.indexedoptionform.IndexedOptionForm(
        index="i32",
        content=ak.forms.emptyform.EmptyForm(),
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "IndexedOptionArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedOptionArray",
            "index": "i32",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "IndexedOptionArray",
        "index": "i32",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedOptionArray",
            "index": "i64",
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "IndexedOptionArray",
        "index": "i64",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "IndexedOptionArray",
            "index": "i32",
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "IndexedOptionArray",
        "index": "i32",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_ByteMaskedForm():
    assert (
        str(
            ak.forms.bytemaskedform.ByteMaskedForm(
                "i8", ak.forms.emptyform.EmptyForm(), True
            )
        )
        == """{
    "class": "ByteMaskedArray",
    "mask": "i8",
    "valid_when": true,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bytemaskedform.ByteMaskedForm(
                "i8", ak.forms.emptyform.EmptyForm(), False
            )
        )
        == """{
    "class": "ByteMaskedArray",
    "mask": "i8",
    "valid_when": false,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bytemaskedform.ByteMaskedForm(
                "i8",
                ak.forms.emptyform.EmptyForm(),
                True,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "ByteMaskedArray",
    "mask": "i8",
    "valid_when": true,
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.bytemaskedform.ByteMaskedForm(
                mask="i8", content=ak.forms.emptyform.EmptyForm(), valid_when=True
            )
        )
        == "ByteMaskedForm('i8', EmptyForm(), True)"
    )
    assert (
        repr(
            ak.forms.bytemaskedform.ByteMaskedForm(
                mask="i8",
                content=ak.forms.emptyform.EmptyForm(),
                valid_when=True,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "ByteMaskedForm('i8', EmptyForm(), True, parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.bytemaskedform.ByteMaskedForm(
        "i8", ak.forms.emptyform.EmptyForm(), True
    ).to_dict(verbose=False) == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "valid_when": True,
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.bytemaskedform.ByteMaskedForm(
        "i8", ak.forms.emptyform.EmptyForm(), True
    ).to_dict() == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "valid_when": True,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.bytemaskedform.ByteMaskedForm(
        mask="i8",
        content=ak.forms.emptyform.EmptyForm(),
        valid_when=True,
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "valid_when": True,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "ByteMaskedArray",
            "mask": "i8",
            "valid_when": True,
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "valid_when": True,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ByteMaskedArray",
            "mask": "i64",
            "valid_when": True,
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "ByteMaskedArray",
        "mask": "i64",
        "valid_when": True,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "ByteMaskedArray",
            "mask": "i8",
            "valid_when": True,
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "ByteMaskedArray",
        "mask": "i8",
        "valid_when": True,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_BitMaskedForm():
    assert (
        str(
            ak.forms.bitmaskedform.BitMaskedForm(
                "u8", ak.forms.emptyform.EmptyForm(), True, True
            )
        )
        == """{
    "class": "BitMaskedArray",
    "mask": "u8",
    "valid_when": true,
    "lsb_order": true,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bitmaskedform.BitMaskedForm(
                "u8", ak.forms.emptyform.EmptyForm(), False, True
            )
        )
        == """{
    "class": "BitMaskedArray",
    "mask": "u8",
    "valid_when": false,
    "lsb_order": true,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bitmaskedform.BitMaskedForm(
                "u8", ak.forms.emptyform.EmptyForm(), True, False
            )
        )
        == """{
    "class": "BitMaskedArray",
    "mask": "u8",
    "valid_when": true,
    "lsb_order": false,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bitmaskedform.BitMaskedForm(
                "u8", ak.forms.emptyform.EmptyForm(), False, False
            )
        )
        == """{
    "class": "BitMaskedArray",
    "mask": "u8",
    "valid_when": false,
    "lsb_order": false,
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.bitmaskedform.BitMaskedForm(
                "u8",
                ak.forms.emptyform.EmptyForm(),
                True,
                False,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "BitMaskedArray",
    "mask": "u8",
    "valid_when": true,
    "lsb_order": false,
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(
            ak.forms.bitmaskedform.BitMaskedForm(
                mask="u8",
                content=ak.forms.emptyform.EmptyForm(),
                valid_when=True,
                lsb_order=False,
            )
        )
        == "BitMaskedForm('u8', EmptyForm(), True, False)"
    )
    assert (
        repr(
            ak.forms.bitmaskedform.BitMaskedForm(
                mask="u8",
                content=ak.forms.emptyform.EmptyForm(),
                valid_when=True,
                lsb_order=False,
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "BitMaskedForm('u8', EmptyForm(), True, False, parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.bitmaskedform.BitMaskedForm(
        "u8", ak.forms.emptyform.EmptyForm(), True, False
    ).to_dict(verbose=False) == {
        "class": "BitMaskedArray",
        "mask": "u8",
        "valid_when": True,
        "lsb_order": False,
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.bitmaskedform.BitMaskedForm(
        "u8", ak.forms.emptyform.EmptyForm(), True, False
    ).to_dict() == {
        "class": "BitMaskedArray",
        "mask": "u8",
        "valid_when": True,
        "lsb_order": False,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.bitmaskedform.BitMaskedForm(
        mask="u8",
        content=ak.forms.emptyform.EmptyForm(),
        valid_when=True,
        lsb_order=False,
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "BitMaskedArray",
        "mask": "u8",
        "valid_when": True,
        "lsb_order": False,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "BitMaskedArray",
            "mask": "u8",
            "valid_when": True,
            "lsb_order": False,
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "BitMaskedArray",
        "mask": "u8",
        "valid_when": True,
        "lsb_order": False,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "BitMaskedArray",
            "mask": "i64",
            "valid_when": True,
            "lsb_order": False,
            "content": {"class": "EmptyArray"},
        }
    ).to_dict() == {
        "class": "BitMaskedArray",
        "mask": "i64",
        "valid_when": True,
        "lsb_order": False,
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "BitMaskedArray",
            "mask": "u8",
            "valid_when": True,
            "lsb_order": False,
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "BitMaskedArray",
        "mask": "u8",
        "valid_when": True,
        "lsb_order": False,
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_UnmaskedForm():
    assert (
        str(ak.forms.unmaskedform.UnmaskedForm(ak.forms.emptyform.EmptyForm()))
        == """{
    "class": "UnmaskedArray",
    "content": {
        "class": "EmptyArray"
    }
}"""
    )
    assert (
        str(
            ak.forms.unmaskedform.UnmaskedForm(
                ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "UnmaskedArray",
    "content": {
        "class": "EmptyArray"
    },
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )
    assert (
        repr(ak.forms.unmaskedform.UnmaskedForm(content=ak.forms.emptyform.EmptyForm()))
        == "UnmaskedForm(EmptyForm())"
    )
    assert (
        repr(
            ak.forms.unmaskedform.UnmaskedForm(
                content=ak.forms.emptyform.EmptyForm(),
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "UnmaskedForm(EmptyForm(), parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.unmaskedform.UnmaskedForm(ak.forms.emptyform.EmptyForm()).to_dict(
        verbose=False
    ) == {
        "class": "UnmaskedArray",
        "content": {"class": "EmptyArray"},
    }
    assert ak.forms.unmaskedform.UnmaskedForm(
        ak.forms.emptyform.EmptyForm()
    ).to_dict() == {
        "class": "UnmaskedArray",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.unmaskedform.UnmaskedForm(
        content=ak.forms.emptyform.EmptyForm(),
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "UnmaskedArray",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {"class": "UnmaskedArray", "content": {"class": "EmptyArray"}}
    ).to_dict() == {
        "class": "UnmaskedArray",
        "content": {
            "class": "EmptyArray",
            "parameters": {},
            "form_key": None,
        },
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "UnmaskedArray",
            "content": {"class": "EmptyArray"},
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "UnmaskedArray",
        "content": {"class": "EmptyArray"},
        "parameters": {"x": 123},
        "form_key": "hello",
    }


def test_UnionForm():
    assert (
        str(
            ak.forms.unionform.UnionForm(
                "i8",
                "i32",
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
            )
        )
        == """{
    "class": "UnionArray",
    "tags": "i8",
    "index": "i32",
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ]
}"""
    )
    assert (
        str(
            ak.forms.unionform.UnionForm(
                "i8",
                "u32",
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
            )
        )
        == """{
    "class": "UnionArray",
    "tags": "i8",
    "index": "u32",
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ]
}"""
    )
    assert (
        str(
            ak.forms.unionform.UnionForm(
                "i8",
                "i64",
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
            )
        )
        == """{
    "class": "UnionArray",
    "tags": "i8",
    "index": "i64",
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ]
}"""
    )
    assert (
        str(
            ak.forms.unionform.UnionForm(
                "i8",
                "i32",
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == """{
    "class": "UnionArray",
    "tags": "i8",
    "index": "i32",
    "contents": [
        {
            "class": "EmptyArray"
        },
        "bool"
    ],
    "parameters": {
        "x": 123
    },
    "form_key": "hello"
}"""
    )

    assert (
        repr(
            ak.forms.unionform.UnionForm(
                "i8",
                "i32",
                [
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
            )
        )
        == "UnionForm('i8', 'i32', [EmptyForm(), NumpyForm('bool')])"
    )
    assert (
        repr(
            ak.forms.unionform.UnionForm(
                tags="i8",
                index="i32",
                contents=[
                    ak.forms.emptyform.EmptyForm(),
                    ak.forms.numpyform.NumpyForm("bool"),
                ],
                parameters={"x": 123},
                form_key="hello",
            )
        )
        == "UnionForm('i8', 'i32', [EmptyForm(), NumpyForm('bool')], parameters={'x': 123}, form_key='hello')"
    )

    assert ak.forms.unionform.UnionForm(
        "i8",
        "i32",
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
    ).to_dict(verbose=False) == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i32",
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
    }
    assert ak.forms.unionform.UnionForm(
        "i8",
        "i32",
        [ak.forms.emptyform.EmptyForm(), ak.forms.numpyform.NumpyForm("bool")],
    ).to_dict() == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i32",
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.unionform.UnionForm(
        tags="i8",
        index="i32",
        contents=[
            ak.forms.emptyform.EmptyForm(),
            ak.forms.numpyform.NumpyForm("bool"),
        ],
        parameters={"x": 123},
        form_key="hello",
    ).to_dict(verbose=False) == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i32",
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }
    assert ak.forms.from_dict(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i32",
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
        }
    ).to_dict() == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i32",
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "u32",
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
        }
    ).to_dict() == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "u32",
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i64",
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
        }
    ).to_dict() == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i64",
        "contents": [
            {
                "class": "EmptyArray",
                "parameters": {},
                "form_key": None,
            },
            {
                "class": "NumpyArray",
                "primitive": "bool",
                "inner_shape": [],
                "parameters": {},
                "form_key": None,
            },
        ],
        "parameters": {},
        "form_key": None,
    }
    assert ak.forms.from_dict(
        {
            "class": "UnionArray",
            "tags": "i8",
            "index": "i32",
            "contents": [
                {"class": "EmptyArray"},
                "bool",
            ],
            "parameters": {"x": 123},
            "form_key": "hello",
        }
    ).to_dict(verbose=False) == {
        "class": "UnionArray",
        "tags": "i8",
        "index": "i32",
        "contents": [
            {"class": "EmptyArray"},
            "bool",
        ],
        "parameters": {"x": 123},
        "form_key": "hello",
    }
